# -*- coding: utf-8 -*-
#
# Copyright (C) 2010-2016 YVertical, PPMessage
# Ding Guijin, guijin.ding@ppmessage.com
# Ning Ben, ben.ning@yvertical.com
#
# All rights reserved
#

from .ppmt import ppmt_main

def mqtt_server():
    ppmt_main()
    return

__version__ = "1.0.4"
